/*
 * InstanceProblem.java
 *
 * Created on 13 mai 2002
 */
 
/** Class InstanceProblem <p>
 * This class allows an instance to be represented and handled.
 * Company:      ONERA/DCSD/CD<p>
 * @author Matthieu Lombard mlombard@operamail.com
 * @version 0.0
 * @see SolutionCheck
 * @see Request
 * @see Strip
 */


import java.io.*;
import java.util.*;

public class InstanceProblem {
  
//-----------------------------------------------------------------------------------
//*******************            A T T R I B U T E S            *********************
//-----------------------------------------------------------------------------------
  
  /** data set number*/
  protected int dataSet;
  /** grid number */
  protected int grid;
  /** track number */
  protected int track;
  
  /** number of requests */
  protected int Nr;
  /** map of the requests indexed by the request index */
  protected static Map mRequest;
  
  /** number of strips*/
  protected int Ns;
  /** list of the strips */
  protected static List lStrip;
  /** map of the strips indexed by the strip index */
  protected static Map mStrip;
  
  /** list of candidate strip acquisitions */
  protected static List lImage;

//-----------------------------------------------------------------------------------
//*******************         C O N S T R U C T O R S            ********************
//----------------------------------------------------------------------------------- 
  /**  Constructor that reads the data from the file fileName
   * @param fileName the name of the file containing the instance data
   */
  public InstanceProblem(String fileName){
    
    initListsAndMaps();
    readInstance(fileName);
    constructListOfStripsPerRequest();
    constructListOfImages();
    
  }//InstanceProblem

//-----------------------------------------------------------------------------------
//*******************              M E T H O D S                 ********************
//----------------------------------------------------------------------------------- 
  //---------------------------------initListsAndMaps---------------------------
  /** initialization of the two lists and two maps of the class <BR>
   */
  protected void initListsAndMaps(){

    mRequest = new HashMap();
    mStrip = new HashMap();
    lStrip = new ArrayList();
    lImage = new ArrayList();

  }//initListsAndMaps

  //-------------------------------readInstance--------------------------------
  /** Method that reads the instance file : <BR>
   *  If the reading is successful, the InstanceProblem object is initialized with 
   *  the read data. <br>
   *  Otherwise, the program stops immediately with an error message.<br>
   */
  public void readInstance(String fileName){
    
    String line;
    StringTokenizer st;
    File f = new File(fileName);

    try {
         System.out.println("\nReading instance file " + fileName + " ...");
         // reading of the file with a BufferedReader
         BufferedReader br = new BufferedReader(new FileReader(f));
         
         //the file is read line after line until the end of file 
         //first line
         line = br.readLine();
         st = new StringTokenizer(line," ");
         dataSet = (Integer.valueOf(st.nextToken())).intValue();
         grid = (Integer.valueOf(st.nextToken())).intValue();
         track = (Integer.valueOf(st.nextToken())).intValue();
         //the program ignores the next two numbers
         st.nextToken();
         st.nextToken();
         
         //second line
         line = br.readLine();
         st = new StringTokenizer(line," ");
         Nr = (Integer.valueOf(st.nextToken())).intValue();
         Ns = (Integer.valueOf(st.nextToken())).intValue();

         //third line
         //the program ignores this line
         line = br.readLine();

         //reading of the requests
         for ( int i = 0 ; i < Nr ; i++){
               Request request = readRequest(br,fileName);
               mRequest.put(request.i,request);
         }

         Strip strip;
 
         //reading of the strips
         for ( int i = 0 ; i < Ns ; i++ ) {
               strip = readStrip(br,fileName);
               lStrip.add(strip);
               mStrip.put(strip.j,strip);
         }         
    }
    
    catch ( Exception e ) {
            System.out.println("*** Error when reading information in the instance file  : " 
                               + fileName);
            e.printStackTrace();
            System.exit(1);
    }

  }//readInstance

 
  //-------------------------------readRequest---------------------------------
  /**  Method that reads the information associated with a request : <BR>
   *   If the reading is successful, the Request object is initialized with the 
   *   read data . <br>
   *   Otherwise, the program stops immediately with an error message.<br>
   */
  protected Request readRequest(BufferedReader br,String fileName){

    Request request = null;

    try {
        String line = br.readLine();
        StringTokenizer st = new StringTokenizer(line," ");

        //request index
        Integer i = Integer.valueOf(st.nextToken());
        //gain associated with the request
        double Gi = (Integer.valueOf(st.nextToken())).doubleValue();
        //surface of the request
        double Si = (Double.valueOf(st.nextToken())).doubleValue();
        //type of the request; either target or polygon
        int Ti = (Integer.valueOf(st.nextToken())).intValue();
        //either mono or stereo request
        boolean Sti = ( (Integer.valueOf(st.nextToken())).intValue() == 1);

        // a stereo request must be acquired twice, so its surface is multiplied by 2
        if ( Sti )
             Si *= 2;

        request = new Request ( i, Gi, Si, Ti, Sti);
    }

    catch ( Exception e ) {
            System.out.println("*** Error when reading requests in the instance file "
                               + fileName);
            e.printStackTrace();
            System.exit(1);
    }

    return request;

  }//readRequest

 //-------------------------------readStrip-----------------------------------
 /**  Method that reads a strip : <BR>
  *   If the reading is successful, the Strip object is initialized with the read data . <br>
  *   Otherwise, the program stops immediately with an error message.<br>
  */
 protected Strip readStrip(BufferedReader br,String fileName){

    Strip strip = null;
    String line;
    StringTokenizer st;

    try{
        //for each strip, the program reads five lines
        //first line
        line = br.readLine();
        st = new StringTokenizer(line," ");
        //index of the strip
        Integer j = Integer.valueOf(st.nextToken());
        //index of the request which the strip comes from        
        Integer Rj = Integer.valueOf(st.nextToken());
        //the program ignores this number
        st.nextToken();
        //index of the twin strip in case of stereo request
        int Twj = (Integer.valueOf(st.nextToken())).intValue();
 
        //second line
        //the program ignores this line
        line = br.readLine();
        
        //third line
        line = br.readLine();
        st = new StringTokenizer(line," ");
        //useful surface of the strip
        double SUj = (Double.valueOf(st.nextToken())).doubleValue();
        //duration of the strip acquisition
        double DUj = (Double.valueOf(st.nextToken())).doubleValue();
 
        //fourth line
        //geographic point 0 = first end of the strip
        GeoPoint geoPoint0 = readGeoPoint(br,fileName);

        //fifth line
        //geographic point 1 = second end of the strip
        GeoPoint geoPoint1 = readGeoPoint(br,fileName);

        // initialization of the computed attributes of the strip
        Request associatedRequest = (Request)mRequest.get(Rj);
 
        strip = new Strip ( j, Rj, Twj, SUj, DUj, geoPoint0, geoPoint1, associatedRequest);
 
    }

    catch ( Exception e ) {
            System.out.println("*** Error when reading strips in the instance file : "
                               + fileName);
            e.printStackTrace();
            System.exit(1);
    }

    return strip;

  }//readStrip

  //--------------------------------------readGeoPoint-------------------------
  /**  Method that reads a geographic point : <BR>
   *   If the reading is successful, the GeoPoint object is initialized with the  
   *   read data. <br>
   *   Otherwise, the program stops immediately with an error message.<br>
   */
  protected GeoPoint readGeoPoint(BufferedReader br,String fileName){

    GeoPoint geoPoint = null;

    try{
        String line = br.readLine();
        StringTokenizer st = new StringTokenizer(line," ");

        //X-ordinate 
        double X = (Integer.valueOf(st.nextToken())).doubleValue();
        //Y-ordinate 
        double Y = (Integer.valueOf(st.nextToken())).doubleValue();
        //earliest date of visibility
        double Te = (Double.valueOf(st.nextToken())).doubleValue();
        //latest date of visibility
        double Tl = (Double.valueOf(st.nextToken())).doubleValue();

        geoPoint =  new GeoPoint(X,Y,Te,Tl);

    }

    catch ( Exception e ) {
            System.out.println("*** Error when reading geographic points in the instance "
                               + "file : " + fileName);
            e.printStackTrace();
            System.exit(1);
    }

    return geoPoint;

  }//readGeoPoint

  //-----------------------------constructListOfStripsPerRequest----------------
  /** Method that assigns to each request the associated list of strips 
   * @see #lStrip
   * @see #mRequest
   * @see Strip#associatedRequest
   */
  protected void constructListOfStripsPerRequest(){
    Strip strip;
    Request request;
    Iterator i = lStrip.iterator();
    while ( i.hasNext()){
            strip = (Strip)i.next();
            request = strip.associatedRequest;
            // adding the strip to the list of strips associated with the request
            request.Stri.add(strip);
    }
  }//constructListOfStripsPerRequest

  //-----------------------------------toString--------------------------------
  public String toString(){
    return "Instance : "
            + "{data set= " + this.dataSet
            + "; grid = "+ this.grid
            + "; track = "+ this.track
            + "; Nr=" + this.Nr
            + "; Ns=" + this.Ns + "}";

  }//toString

 //-----------------------------------lImageadd---------------------------------
 /** Method that adds the two possible strip acquisitions associated with a 
  * strip to the list lImage
  * at the index k. <BR>
  * @param strip strip the two possible strip acquisitions of which will be added to lImage
  * @param k index of the first strip acquisition in lImage
  * @see #lImage
  * @see Image
  * @see Strip */
 protected void lImageadd(Strip strip, int k){

   Image image1,image2;
   image1 = new Image(strip,0,k);
   image2 = new Image(strip,1,k + 1);
   //the two strip acquisitions associated with this strip 
   strip.cImage[0] = image1;
   strip.cImage[1] = image2;
   lImage.add(k, image1);
   lImage.add(k + 1, image2);
   
 }//lImageadd

  //----------------------------------constructListOfImages---------------------------
  /** Initialization method of the list lImage.<BR>
   * It uses the list of strips lStrip and creates lImage in the same order.
   * @see #lImage
   * @see #lImageadd
   */
  protected void constructListOfImages(){
    lImage.clear();
    int k = 0;
    Iterator i = lStrip.iterator();

    while ( i.hasNext() ) {
            lImageadd((Strip)i.next(),k);
            k+=2;
    }

  }//constructListOfImages
  
}// InstanceProblem.java

