/*
 * Solution.java
 *
 * Created on 14 mai 2002
 */
 
/** Class Solution <p>
 * This class reads the solution file. Then it creates the solution list. <br>
 * Then it tests whether this solution is feasible. If so, it computes its gain. 
 * If not, it informs the user that it is not feasible.
 * Company:      ONERA/DCSD/CD<p>
 * @author Matthieu Lombard mlombard@operamail.com
 * @version 0.0
 * @see SolutionCheck
 */


import java.io.*;
import java.util.*;

public class Solution {

//-----------------------------------------------------------------------------------
//*******************            A T T R I B U T E S            *********************
//-----------------------------------------------------------------------------------
  //read attributes
  /** data set number */
  protected int dataSet;
  /** grid number */
  protected int grid;
  /** track number */
  protected int track;
  /** number of selected strip acquisitions */
  protected int Na;
  
  //computed attributes
  /** if the solution is feasible, this attribute equals true;
   *  otherwise, it equals false  */  
  private boolean isFeasible;
  /** list of the strip acquisitions in the solution*/
  private List solutionList;
  /** gain of the solution, if feasible*/
  private double gain;
  /** name of the result file*/
  private String resultFileName;

//-----------------------------------------------------------------------------------
//*******************         C O N S T R U C T O R S            ********************
//----------------------------------------------------------------------------------- 
  public Solution ( String fileName, 
                    String resultFileName) {
      
    this.resultFileName = resultFileName;
    this.solutionList = constructSolutionList(fileName);
    this.isFeasible = verifySolution();
    
    if ( this.isFeasible ) {
         this.gain = computeGain();
    }
    
    displaySolution();
    writeResultsFile(resultFileName);
  }

//-----------------------------------------------------------------------------------
//*******************              M E T H O D S                 ********************
//----------------------------------------------------------------------------------- 

//READING SOLUTION ##################################################################
  //------------------------constructSolutionList-------------------------------
  /** Method that constructs the list of selected strip acquisitions from the 
   *  read data.
   * @param fileName the name of the solution file 
   * @return the list of strip acquisitions in the solution
   */
  public List constructSolutionList ( String fileName ) {
      
    String line;
    StringTokenizer st;
    List listOfSolutionImages = new ArrayList();

    File f = new File(fileName);

    try {
         System.out.println("\nReading solution file " + fileName + " ...");
         // reading of the file with a BufferedReader
         BufferedReader br = new BufferedReader(new FileReader(f));
         //the file is read line after line until the end of file 

         //first line
         line = br.readLine();
         st = new StringTokenizer(line," ");
         dataSet = (Integer.valueOf(st.nextToken())).intValue();
         grid = (Integer.valueOf(st.nextToken())).intValue();
         track = (Integer.valueOf(st.nextToken())).intValue();

         //second line
         line = br.readLine();
         st = new StringTokenizer(line," ");
         //number of selected strip acquisitions
         Na = (Integer.valueOf(st.nextToken())).intValue();

         //reading of the selected strip acquisitions
         for ( int i = 0 ; i < Na ; i++){
               Image image = readImage(br,fileName);
               listOfSolutionImages.add(image);
         }
    }

    catch ( Exception e ) {
            System.out.println("*** Error when reading the solution file : "
                               + fileName);
            e.printStackTrace();
            System.exit(1);      
    }

    return listOfSolutionImages;
    
  }//constructSolutionList
  
  //-------------------------------readImage------------------------------------
  /**  Method that reads a selected strip acquisition: <BR>
   *   if the reading is successful, the Image object is initialized with the 
   *   read data , <br>
   *   otherwise, the program stops immediately with an error message.<br>
   */
  private Image readImage(BufferedReader br,String fileName){

    Strip strip;
    Image image = null;

    String line;
    StringTokenizer st;

    try{
        //for each strip, the program reads one line
        line = br.readLine();
        st = new StringTokenizer(line," ");
        //index of the strip
        Integer Strak = Integer.valueOf(st.nextToken());
        //acquisition direction of the strip
        int Dirak = Integer.valueOf(st.nextToken()).intValue();

        //getting the image in the list lImage
        strip = (Strip)InstanceProblem.mStrip.get(Strak);
        image = strip.cImage[Dirak];
    }

    catch ( Exception e ) {
            System.out.println("*** Error when reading selected strip acquisitions in" 
                               + " the solution file : "
                               + fileName);
            e.printStackTrace();
            System.exit(1);
    }

    return image;

  }//readImage    


//CHECKING SOLUTION FEASIBILITY ####################################################
  //-------------------------------verifySolution-------------------------------
  /**  Method that checks the feasibility of the solution.<br>
   * !!! the solution list must be not empty.<br>
   */
  private boolean verifySolution(){

    Iterator i = solutionList.iterator();
    Image image, inverseImage, twinImage, currentImage, previousImage;
    Strip twinStrip,associatedStrip;

    //checking whether the list of strip acquisitions in the solution is empty or not
    if ( solutionList == null ) {
         System.out.println("*** Error in verifySolution : the list of selected strip"
                            + " acquisitions is empty");
         System.exit(1);
    }

    while ( i.hasNext() ) {
            image = (Image)i.next();
                        
            //checking that the visibility window of each selected strip acquisition is not empty
            if ( image.Tmink > image.Tmaxk ) {
                 System.out.println("*** Error in verifySolution : empty visibility window for "
                                     + " acquisition of strip :" + image.Strak.j);
                 return false;
            }
            
            //checking that no strip is acquired twice
            inverseImage = image.Strak.cImage[( image.Dirak + 1 ) % 2];
            if ( solutionList.contains(inverseImage) ) {
                 System.out.println("*** Error in verifySolution : strip " + image.Strak.j
                                    + " acquired twice");
                 return false;
            }
            
            //checking , for each pair of twin strips associated with a stereo request, 
            // that none or both are acquired and that, if they are both acquired,  
            // their acquisition direction is the same
            if ( image.Strak.Twj != 0 ) {
                 twinStrip = (Strip)InstanceProblem.mStrip.get(new Integer(image.Strak.Twj));
                 twinImage = twinStrip.cImage[image.Dirak];
                 if ( !solutionList.contains(twinImage) ) {
                      System.out.println("*** Error in verifySolution : stereo acquisition constraint "
                                         + " violation for strip :" + image.Strak.j);
                      return false;
                 }
            }
    }
    
    //forward propagation of the temporal constraints
    if ( solutionList.size() > 1 ) {
         
         //checking dates starting from the first selected strip acquisition
         i = solutionList.iterator();
         //current selected strip acquisition
         currentImage = (Image)i.next();
         //earliest starting time of the current selected strip acquisition
         double tmin1k = currentImage.Tmink;

         while ( i.hasNext() ) {
                 previousImage = currentImage;
                 currentImage = (Image)i.next();
                 
                 double DUj = previousImage.Strak.DUj;
                 double DikkPrime = transitionTime(previousImage,currentImage);
                 double tminPrime = tmin1k + DUj + DikkPrime;

                 if ( tminPrime > currentImage.Tmaxk ) {
                      System.out.println("*** Error in verifySolution : empty visibility window for "
                                     + " acquisition of strip :" 
                                     + currentImage.Strak.j
                                     + " after propagation of the temporal constraints");
                      return false;
                 }
                 //earliest starting time of the current selected strip acquisition
                 tmin1k = Math.max( currentImage.Tmink, tminPrime);
         }
         
    }
       
    return true;

  }//verifySolution    

  //-------------------------------transitionTime-------------------------------
  /**  Method that computes the transition time between two strip 
   * acquisitions k and k'.<br>
   */
  private double transitionTime(Image k, Image kPrime) {

   //difference between X-ordinates of the end of k and the beginning of k'
   double dX = ((k.geoPointOutput.X) - (kPrime.geoPointInput.X));
   
   //difference between Y-ordinates of the end of k and the beginning of k'
   double dY = ((k.geoPointOutput.Y) - (kPrime.geoPointInput.Y));                
                
   //ground distance between the end of k and the beginning of k'
   double DikkPrime = Math.sqrt(dX*dX + dY*dY);   
   
   //approximation of the rotation angle the satellite has to perform between 
   //the end of k and the beginning of k'
   double ROkkPrime = 2*Math.atan( DikkPrime/(2*Constants.Hs));
   
   //approximation of the transition time between the end of k and the beginning of k'
   return ( Constants.Dmin + (ROkkPrime/Constants.Vr) );

  }//transitionTime    
  
//COMPUTING SOLUTION GAIN ###########################################################
  //-------------------------------computeGain----------------------------------
  /**  Method that computes the gain of the solution. This gain is the sum of 
   *  the gains of the strip acquisitions in the solution list.<br>
   *  It takes into account partial gains from the strip acquisitions belonging to 
   *  a polygon. <br>
   * @see #fPartialGain
   * @see #solutionList
   * @see Strip#fr
   * @see Strip#associatedRequest
   * @see Request#acquisitionLevel
   * @see Request#totalGain 
   */
  private double computeGain(){

    this.gain = 0.0;
    Strip strip;
    Request request;
    Iterator i = this.solutionList.iterator();
    
    //loop on the list of selected strip acquisitions
    while ( i.hasNext() ) {
            strip = ((Image)i.next()).Strak;
            request = strip.associatedRequest;

            if ( request.Ti == 0 ) {//target
                 request.acquisitionLevel = 1;
            }
            else if ( request.Ti == 2 ) {//polygon
	           request.acquisitionLevel += strip.fr;
      }
    }

    i = ((Collection)InstanceProblem.mRequest.values()).iterator();

    //loop on list of requests
    while ( i.hasNext() ) {
            request = (Request)i.next();
            this.gain += request.totalGain * fPartialGain(request.acquisitionLevel);
    }

    return this.gain;

  }//computeGain    

  //-----------------------------------fPartialGain-----------------------------
  /**  Definition of the breakpoints and of the slopes of the piece-wise linear  
   * function, used to give a higher importance to polygon terminaison
   */
  protected double fPartialGain(double x){
    double a1 = 0.4;
    double a2 = 0.7;
    double p1 = 0.25;
    double p2 = 1;
    double p3 = 2;
    return fPieceWiseLinear(a1,a2,p1,p2,p3,x);
  }//fPartialGain

  //---------------------------------------fPieceWiseLinear---------------------
  /** Definition of the piece-wise linear convex function
   */
  private double fPieceWiseLinear(double a1, double a2,
			   double p1, double p2, double p3,
			   double x){
    // computing error (because double are used)
    double Eps = 0.0001;
    double y = 0;
    
    if ( (x<(0-Eps)) || (x>(1+Eps)) ) {
         System.out.println("*** Error in the function fPieceWiseLinear : acquisition level"
                            + " (=" + x + ") must be a number between 0 and 1");
         System.exit(1);
    }
    
    if( (a1*(p1-p2) + a2*(p2-p3) + p3 > 1+Eps)
	|| (a1*(p1-p2) + a2*(p2-p3) + p3 < 1-Eps) ){
      System.out.println("*** Error in the function fPieceWiseLinear : "
			 + " breakpoints are incompatible");
      System.exit(1);
    }

    if ( x <= a1 )
         y = p1 * x;
    else if ( x <= a2 )
         y = p1*a1 + p2 * (x - a1);
    else if ( x > a2 )
         y = p1*a1 + p2*(a2-a1) + p3 * (x-a2);
    return y ;
    
  }//fPieceWiseLinear
  

//DISPLAYING SOLUTION ###############################################################
  //-------------------------------displaySolution------------------------------
  /**  Method that displays the information about the solution. I<br>
   */
  private void displaySolution() {
   
    StringBuffer s = new StringBuffer();;
    s.append( this.toString() + "\n");

    if ( this.isFeasible ) {
         s.append("           Feasible; g= " + (int)this.gain);
    }
    else 
         s.append("  not feasible !");
    
    System.out.println(s.toString());

  }//displaySolution    

  //-----------------------------------toString---------------------------------
  public String toString(){
    return "Solution : "
            + "{data set= " + this.dataSet
            + "; grid = "+ this.grid
            + "; track = "+ this.track
            + "; Na=" + this.Na + "}";

  }//toString

  //------------------------writeResultsFile-------------------------------
  /** Method that writes the results of the solution checking in the 
   *  result file
   */
  public void writeResultsFile (String fileName) {


    try {
         System.out.println("\nWriting solution checking results in file "
                            + fileName + " ...");

         PrintWriter pw = new PrintWriter(
                              new BufferedWriter(
                                  new FileWriter(fileName) )); 

         //writing instance information
         pw.println( dataSet + " " + grid + " " + track);
         if ( isFeasible ) {
              pw.println( (int)gain );
         }
         else
              pw.println( 0 );
                  
         pw.close();
         System.out.println("File "+ fileName + " generated");
    }

    catch ( Exception e ) {
            System.out.println("*** Error when writing file  : " + fileName);
            e.printStackTrace();
            System.exit(1);
    }
  }//writeResultsFile

}//Solution.java

